/* APE (Allelic Path Explorer) v.1.0
   searches for a consistent genotypic configuration 
   when the known pedigree structure 
   and genetic data for some individuals are given. 


   Copyright (C) Matti Pirinen
   
   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   version 2 as published by the Free Software Foundation.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program (gpl.txt); if not, write to the Free Software
   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

   www.gnu.org/licenses/gpl.html
*/


/* 
   Matti Pirinen Dec-2005: matti.pirinen@helsinki.fi

   Tested under Linux with gcc version 2.95.4 20011002
   to compile: gcc -Wall -O2 APE.c -o APE -lm

   Input and output are explained at the end of this file. 
*/


#define SAMPLE_UNDEFINED 0
// Whether to sample the undefined genes (=1) or leave them as zeroes (=0).
#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include <signal.h>
#include <limits.h>
#include "APE.h"


int main(int argc, char* argv[])
{

  char pedfile[40],datafile[40],afreqfile[40];
  int i,g,marker,j;
  int *knowndata;
  int **fixed_parents,*sexes;
  COMP *C;
  g_list *G;
  pair **genes, *tgenes,**grandp_origins;
  double ***conditionalF,***kinship,**afreq,*ratio_known;
  int *levels;
  int allowed_jumps; 
  //CHILDREN *offspring;
  PARENTAL *parental;
  UL **inconsistencies;
  clock_t kello2;

  kello=clock();
  signal(SIGINT, sigproc);
  srand( (unsigned int) time(NULL) );


  /*** Initializing the known pedigree/data structure. ***/

  printf("\n*** APE ***\n");
  if(argc>1)
    {
      FILE *args;

      if((args= fopen(argv[1],"r"))==NULL)
	{
	  fprintf(stderr, "\nError: Cannot open parameters file \"%s\"!\n",argv[1]);
	  fprintf(stderr,"Correct usage:\"APE parmeters_file\"\nSee README for details about parameters.\n\n");
	  exit(3);
	}

      read_parameters(args,pedfile,datafile,afreqfile,&N_markers, &M_comp, &M_lev, &kappa, &diff_exp, &M_clu, &allowed_jumps);
      fclose(args);
    }
  else
    { 
      fprintf(stderr, "\nError: No parameters file specified!\nCorrect usage:\"APE parmeters_file\"\nSee README for details.\n\n");
      exit(3);
    }

  //times=fopen("times","a"); // This file is for test runs only.
  //fprintf(times,"A");

 /* reads the known pedigree from pedfile */

  { 
    FILE *ped;
    if((ped= fopen(pedfile,"r"))==NULL)
      {
	fprintf(stderr, "\nError: Cannot open pedigree file \"%s\"!\nSee README for details about parameters.\n\n",pedfile);
	exit(3);
      }
    fixed_parents = readPedigree(ped,&N_individuals);
    fclose(ped); 

    ped = fopen(pedfile,"r");
    sexes = read_sexes(ped);
    fclose(ped);
  }  

  //offspring = list_children(fixed_parents);
  //print_offspring(offspring,stdout);

  //fprintf(times,"%d\n",N_individuals); //this is for test runs only

  if( (genes = (pair **) malloc(N_individuals * sizeof(pair*)) ) == NULL)
    alloc_error("main with 'genes'");

  for(i=0; i<N_individuals ;i++)
    {
      if( (genes[i] = (pair *) calloc(N_markers, sizeof(pair)) ) == NULL)
	alloc_error("main with 'genes'");
    } 


 if( (tgenes = (pair *) malloc(N_individuals * sizeof(pair)) ) == NULL)
   alloc_error("main with 'tgenes'");

 { 
   FILE *data;
   if((data= fopen(datafile,"r"))==NULL)
     {
       fprintf(stderr, "\nError: Cannot open data file \"%s\"!\n",datafile);
       fprintf(stderr,"See README for details about parameters.\n\n"); 
       exit(3);
     }
   knowndata = readData(data,genes,&N_knowndata);   
   fclose(data);
   infer_from_homozygosity(fixed_parents,genes); 

   if((data= fopen(afreqfile,"r"))==NULL)
     {
       fprintf(stderr, "\nError: Cannot open allele frequency file \"%s\"!\n",afreqfile);
       fprintf(stderr,"See README for details about parameters.\n\n");
       exit(3);
     }
   afreq = readAfreq(data);
   fclose(data);
 }
 
 printf("Pedigree has %d individuals ",N_individuals);
 printf("of which %d has some observed data.\n",N_knowndata);
 printf("The fraction is %5.3f.\n",1.0*N_knowndata/N_individuals);

 if( (levels = (int*) malloc(N_individuals*sizeof(int))) == NULL)
   alloc_error("main with levels");

  for(i=0;i<N_individuals;i++)
    levels[i]=-1;

  N_levels=0;
  
  for(i=0;i<N_individuals;i++)
   {
     count_level(i,fixed_parents,levels);
     N_levels=(levels[i]+1>N_levels) ? levels[i]+1 : N_levels;
   }
  
  if ((C = (COMP *) malloc(N_levels * sizeof(COMP)) ) == NULL)
    alloc_error("main with 'C'");
  
  make_components(C,fixed_parents,knowndata,levels);
  parental = find_parental_info(C);  
  //print_parental_info(stdout,parental);

  /*{
    FILE *outp;
    outp=fopen(COMPONENTS_OUTPUT,"w");
    print_components(C,outp);
    //print_components_info(C,outp);
    //print_pedigree_info(outp,C,sexes);    
    fclose(outp);
    }*/

  printf("Counting kinship...\n");
  //kinship=count_kinship_with_recursion(C,fixed_parents,levels);
  //kinship=count_kinship(C,fixed_parents);
  kinship=count_kinship2(C,fixed_parents,parental);
  printf("...kinship ready.\n");

  /*{
    FILE *outp;
    outp=fopen("kinship","a");
    print_kinship(kinship,C,outp);
    fclose(outp);
  }*/
  find_component_ordering(C,kinship);

  if( (G = (g_list *) malloc((N_levels-1) * sizeof(g_list)) ) == NULL)
    alloc_error("main with 'G'");

  if( (inconsistencies=(UL**) malloc(N_levels*sizeof(UL*))) == NULL)
    alloc_error("main with inconsistencies");

  for(i=0;i<N_levels;i++)
    if( (inconsistencies[i] = (UL*) calloc(C[i].ncomponents,sizeof(UL))) == NULL)
      alloc_error("main with inconsistencies");

 /************************ Checking all possible configurations. ************************/

  printf("\nSearching for descent paths...\n");
  for(marker=0;marker<N_markers;marker++)
    {

      int new, inconsistent_comp;
      unsigned long jump_back=0;

      /* Counting the pseudo-conditional probabilities to conditionalF. */
      conditionalF = transmission_probabilities(genes, fixed_parents, C, marker, N_alleles[marker], afreq[marker]);
      
      check_consistency_of_components(C,N_levels,fixed_parents,genes,marker,conditionalF);
      
      /*{
	FILE *outp;
	outp=fopen("conditional_probabilities","w"); 
	//fprintf(outp,"%d %d\n",n_infer[0],n_infer[1]);
        print_conditionalF(conditionalF, C,outp,N_alleles[marker]);
	fclose(outp);
	}*/
      ratio_known=count_known_data(genes,marker,C);

      {
	//FILE *outp;
	//outp=fopen("cen","w");
      
      if(marker>0)
	{
	  G[N_levels-2].n=0; // Since this was not ordered during the previous round.
	  for(i=0;i<N_levels-1;i++)
	    free_glist(&G[i],C[i].ncomponents);
	}

      for(i=0; i<N_levels-1; i++)
	G[i].prev_config=0;

      for(i=0; i<N_individuals;i++)
	{
	  tgenes[i][0]=genes[i][marker][0];
	  tgenes[i][1]=genes[i][marker][1];
	}

      g=0; //The algorithm of APE begins, g=current level

      do
	{

	  inconsistent_comp=find_configurations(G, C, g, fixed_parents, genes, marker, conditionalF);

	  if(inconsistent_comp == -1) //found consistent configuration(s) for level g+1!
	    {
	      printf("%d: ",g);
	      if(g>0)
		update_tgenes(tgenes,g,C,G,1); 
	      //Updates temporary configuration to the level g 
	      //i.e. current configurations at levels 0,...,g-1 are fixed. 
	      if(g<N_levels-2)
		{
		  //constrain_components(G,C,g);
		  //order_generation2(G,C,g,kinship,tgenes,fixed_parents,offspring,levels);
		  order_generation(G,C,g,kinship,tgenes,ratio_known);
		  //print_combinations(G[g].ordered,G[g].n,C[g].ncomponents,g,outp);
		}
	      else //We have reached the eldest level.
		build_pedigree(genes,G,C,marker);

	      //print_generation(G,C,g,outp);
	      //print_genotypes(G,C,g,outp);
	      ++g;
	    }
	  else //No consistent configuration found for level g+1.
	    {
	      //fprintf(outp,"inconsistent component at %d: %d \n",g,inconsistent_comp);
	      inconsistencies[g][inconsistent_comp]++;
	      no_cons++;

	      if(! ( new= (++G[g].prev_config != G[g-1].n) ) )
		{
		  G[g].prev_config=0;
		  kello2=clock();
		  if( ((double) (kello2-kello) / (double) CLOCKS_PER_SEC)>100)
		    {
		      sek+=(double)(kello2-kello)/(double) CLOCKS_PER_SEC;
		      kello=kello2;
		    }
		}

	      while(!new && g > 0)
		{
		  jump_back++;
		  if( allowed_jumps > 0 && jump_back >= (unsigned) allowed_jumps )
		    {
		      printf("\nProgram aborted since %ld jumps were encountered.\n",jump_back); 
		       {
			 FILE *out=fopen(EXTRACTIONS_FILE,"w");
			 print_inconsistencies(out,inconsistencies,C,no_cons,inconsistency_frac);
			 fclose(out);

			 print_connected_components(fixed_parents);
		       }
		      print_comments(2);
		      exit(2);
		    }
		  --g;
		  free_glist(&G[g],C[g].ncomponents);
		  if(g>0 && (++G[g].prev_config==G[g-1].n))
		    G[g].prev_config=0;

		  update_tgenes(tgenes,g,C,G,-1);
		  
		  new=(G[g].prev_config>0);
		}
	      
	      
	    } 
	      
	}while(g>0 && g < N_levels-1);
	    
      //fclose(outp);
      
      if(g==0)
	{
	  printf("Unable to find a configuration for marker %d.\n",marker+1);
	  printf("unsuccessful configurations=%ld; backward jumps=%ld\n", no_cons, jump_back);
	  {
	    FILE *out=fopen(EXTRACTIONS_FILE,"w");
	    print_inconsistencies(out,inconsistencies,C,no_cons,inconsistency_frac);
	    fclose(out);	    

	    print_connected_components(fixed_parents);
	  }
	  print_comments(1);
	  exit(2);
	}
      }
      printf("A consistent configuration found for marker %d. \n",marker+1);
      printf("unsuccessful configurations=%ld; backward jumps=%ld\n", no_cons, jump_back);

      if(N_markers>0)
	{ //A different file should be used for each marker!!!! (in future versions)
	  FILE *out=fopen(SIMPLE_OUTPUT,"w");
	  print_pedigree(pedfile,genes,marker,out);
	  fclose(out);
	}

      for(i=0;i<N_individuals;i++)
	{
	  for(j=0;j<N_alleles[marker];j++)
	    free(conditionalF[i][j]);
	  free(conditionalF[i]);
	}
      free(conditionalF);
      free(ratio_known);

    }	     /* for(marker=0;marker<N_markers)*/

  if(!check_consistency(genes, fixed_parents)) // with Mendel
    exit(5);

  {
    FILE *data=fopen(datafile,"r");
    if(!check_compatibility(data,genes)) // with the datafile
      exit(5);
    fclose(data);
  }

  {
    FILE *fout;
    /*fopen(SIMPLE_OUTPUT,"w");
    print_pedigree(pedfile,genes,N_markers-1,fout);
    fclose(fout);*/

    grandp_origins = find_grandp_origins(genes, fixed_parents);
    total_configuration(genes,grandp_origins,fixed_parents,C,afreq,SAMPLE_UNDEFINED);

    fout = fopen(TOTAL_OUTPUT,"w");
    print_total(pedfile,genes,grandp_origins,levels,fout);
    fclose(fout);
    
    print_connected_components(fixed_parents);

  }

  //draw_pedigree("pedfile");
  //print_defined_genes(genes,C);
  print_comments(0);


  return 0;

}

#include "APE_initialize.c"
#include "APE_general.c"
#include "APE_order.c"
#include "APE_print.c"


/* 

To compile APE (with GNU C-compiler):
gcc -Wall -O2 APE.c -o APE -lm

To use APE:
./APE parameters_file


parameters
-----------

Parameters are in a file that is given as an argument to the program. File should look like this:
------------------------------------------------------------------------------------------------
pedfile
datafile
afreqfile

_markers:
3
_alleles:
3,4,3
_M_comp:
0
_M_lev:
0
_kappa: 
0.1
_diff_exp:
1.0
_M_clu:
10
_allowed:
0

Do not change the titles that begin with underscores.

The three files specified are

the pedigree file
the genotype data file
the allelic frequency file

Their formats are explained below.
The numerical values stand for

_markers = number of marker loci
_alleles = maximum number of alleles for each locus (the maximum label of alleles found in the datafile, should be a positive integer).
           If there are many loci, numbers should be separated by commas. 
_M_comp = maximum number of configurations in a single component. 
          (0 means no restrictions)
_M_lev = number of saved configurations (M_lev of the most probable configurations are saved at each level). 
         (0 means no restrictions)
_kappa = parameter which tells how probabilities and variation parameters are united to scores. 
         (0.0 means that only probabilities count.)
_diff_exp = exponent that is used in counting differences between genotypes 
            (default value 1.0 makes APE faster)
_M_clu = maximum number of configurations at cluster. 
         (0 means no restrictions.)
_allowed = maximum number of backward jumps that are allowed before the program stops.
           (0 means no restrictions)


Input files: 
-------------

There should be three input files whose names are given in the parameters file.
Namely 'pedfile', 'datafile' and 'afreqfile'. 

'pedfile' 

On the first row there should be the number of individuals (say n) in the pedigree.
Then n lines follow, one for each individual, specifying

index  (father's index)  (mother's index)  sex

where indexing starts from 1. 
Sexes are 1 = male, 2 = female. 
For founders father and mother should be 0 and every individual has either two or zero nonzero parents.
Individuals' indexes should be in the range  [1...n], parents' indexes in the range [0...n].
Order of individuals is free but every individual (from 1 to n) should be entered once and only once.

e.g. pedigree with 6 individuals and 3 founders (4,5,6)

6
1 4 6 1 
2 5 6 1
3 4 6 2
4 0 0 1
5 0 0 1
6 0 0 2


'datafile'

Datafile begins with the number of (partially) observed genotypes.
For each individual whose genotype is observed there should be the following 3 lines

index
first set of alleles for all markers (N_markers entries)
second set of alleles for all markers (N_markers entries)

('set of alleles' could be haplotypes, if we only knew them)

Zero stands for unobserved allele.
Allowed labels for alleles are thus integers from 0 to N_alleles[locus].


e.g (partially observed genotypes= 2, N_markers = 3, N_alleles=9,8,9)

2

1
3 8 4
9 5 6

2
4 5 1
3 7 9


(blank lines are not obligatory)



'afreq file'

This file contains allelic frequenies.
For each marker it contains frequencies in {}-brackets.
For example, if N_markers=2 and markers contain 3 and 4 alleles, respectively, we could have a file like

{0.3,0.4,0.3}
{0.1,0.4,0.1,0.4}

For each locus the sum of frequencies should be 1.0+/- 0.01.



Output:
******

Output is given in two files. (Names of the files are macros defined in APE.h)

paths.out (macro SIMPLE_OUTPUT)
*********

paths.out is a simpler file defining only the genotypes of those individuals that are part
of the descent paths of the observed alleles. It contains 3 lines for each individual.

index (father's ind.) (mother's ind.) sex 
first series of alleles (N_markers entries)
second series of alleles (N_markers entries) 

Note that series here are NOT haplotypes i.e. both series may contain alleles from both father and mother!
(Phases are assigned in function find_grandp_origins when it is called for making total configuration (see below).)

total.out (macro TOTAL_OUTPUT)
********

total.out contains a total configuration for the pedigree and the data.
There are two options controlled by the macro SAMPLE_UNDEFINED.
If SAMPLE_UNDEFINED == 1 then the undefined genes are sampled for founders and dropped down through the pedigree 
If SAMPLE_UNDEFINED == 0 then the undefined alleles remain zeroes in the total.out.
Phase information is included as haplotypes and grandparental origins are found out.

In total.out there are 5 lines for each individual.

index (father's ind.) (mother's ind.) sex
the haplotype inherited from the father (N_markers entries)
grandparental origins of each allele in the paternal haplotype (0 = from the grandmother, 1 = from the grandfather)
the haplotype inherited from the mother (N_markers entries)
grandparental origins of each allele in the maternal haplotype (0 = from the grandmother, 1 = from the grandfather)



exit codes of APE
**********
0: a consistent configuration has been found (normal for succesful attempts)
1: an error in dynamic memory allocation
2: a consistent configuration not found (normal for unsuccesful attempts)
3: problems reading parameters/datafiles or inconsistency with the given data and the given pedigree
4: aborted by the user with Ctrl-C
5: inconsistency with data or Mendelian laws, POSSIBLY A PROGRAMMING ERROR!!!!
6: other reasons

NOTE: the exit code can be 2 even if the data and the pedigree are inconsistent, since
we can not check all possibilities in large pedigrees.


-------------------------------------------------

EXTRACTING COMPONENTS

-------------------------------------------------

Currently the extraction of components is done with a different APE called APEe (APE extraction).

To compile APEe  
gcc -Wall -O2 APE_extract.c -o APEe -lm

To run APEe 
./APEe parameters_file

The parameters are exactly the same as with APE (see above).
There are two files that are specific for APEe, however.

The file
'extracted_components'

contains the names of those components that APEe shall extract.
The names are the level and the ordinal number of the component on that level.
The program checks that the components are valid with repect to the given pedigree.

An example of the file 'extracted component' (macro in APE.h) is

10 54
9 85
10 38

which result in the extraction of the three components: 54 and 38 on the 10th level
and 85 on the 9th level. Levels start from 0.

When extracting components, APEe sets everything outside the components zero. If
it is successful with that restricted data, it checks whether it has extended this 
partial data in such a way that some new data on someone not previously genotyped are
achieved. It then fixes some of these new genotypes and writes the new datafile
'extracted_data' (macro in APE.h). The new datafile contains a copy of the datafile
that APEe was given in the parameters file as well as some of the new genotypes.
APEe also writes the new genotypes on the screen.

Note that after the extraction the basic APE as well as subsequent extractions with APEe
must be run on the new datafile. However, sometimes extractions result in inconsistencies
with the data outside the extracted components and then the procedure must be started
from the beginning with the original datafile. It is also possible to run APEe
many times if it seems that not much data is fixed after the first run.

Both APE and APEe automatically write the hard components to 'extracted_components',
if they don't succeed in finding a configuration. Sometimes, however, one needs to extract
more components jointly and then one has to modify the file 'extracted_component' manually.

The ratio of inconsistencies needed in order that a component will be printed out to 
'extracted_component' is defined in the file APE.h (and is by default 0.1). 
 
*/
