/*
  APE v.1.0. Copyright (C) Matti Pirinen
   
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  version 2 as published by the Free Software Foundation.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program (gpl.txt); if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  
  02110-1301, USA.
  
  www.gnu.org/licenses/gpl.html
*/


/* FILE: APE_general.c
   Contains general functions of APE.
   See APE.h for list of functions.
*/



void alloc_error(char *c)
{
  fprintf(stderr,"ERROR allocating memory in function:\n %s\n",c);
  fprintf(stderr,"Program exits!\n");
  exit(1);
}

int compatible(int a, int parent[2])
{
  
  /*Tries to update parent[2] so that the allele "a" may have been inherited from the parent. 
    If this is not possible: returns 0.
    If there's no need for updating: returns 1.
    If the allele a is added to the genotype: returns 2. */

  if(parent[0]==a || parent[1]==a || a==0)
    return 1;
  if(parent[0]==0 || parent[1]==0)
    {
      if(parent[0]==0)
	parent[0]=a;
      else
	parent[1]=a;
      return 2;
    }
  return 0;
}




double conditional_probability(int i,int genes[2],double ***conditionalF,int n_alleles)
{

  /*Returns the conditional probability based on the observed alleles above individual i
    that he/she has the genotype genes[2].
  */

  int a;

  int nzeros=(genes[0] == 0) + (genes[1] == 0);

  switch(nzeros)
    {
    case 0:

      return conditionalF[i][genes[0]-1][genes[1]-1];
      break;

    case 1:
      
      a=(genes[0]>0) ? genes[0]-1 : genes[1]-1;

      return conditionalF[i][a][n_alleles];
      break;

    case 2:

      return 1.0;
      break;
    }


  return 0.0;
}


int contains(int a, pair gene)
{
  /*Can allele 'a' be transmitted by individual with genotype 'gene'? */

  if(a == 0 || a==gene[0] || a==gene[1])
    return 1;
  else 
    return 0;
}



void delete_combination(pair **ttensor, int *i, int *ntensor, int *ntensor0)
{
  /* Deletes ttensor[*i] and
     ppdates *i and *ntensor0 and *ntensor appropriately for update_list(...) function.
  */

  int j=*i;

  free(ttensor[j]);
  
  if(*ntensor0 - 1 > j)
    {
      ttensor[j]=ttensor[*ntensor0-1];
      --(*ntensor0);
      --(*i);
      j=*ntensor0;
    }
  
  if((*ntensor)-1 > j)
    ttensor[j]=ttensor[*ntensor-1];

  --(*ntensor);
  
}

void delete_combination2(pair **ttensor,int *i, int *ntensor)
{
  /*Deletes element ttensor[i] so that remaining elements
    are at the places 0,...,ntensor-2.
    Reduces i and ntensor by one. (i may be -1 afterwards!)*/

 int j=*i;
 free(ttensor[j]);

 if((*ntensor)-1 > j)
   ttensor[j]=ttensor[*ntensor-1];

 --(*i);
 --(*ntensor);

}


int find_configurations(g_list *G, COMP *C, int g, int **fixed_parents, pair** genes, int marker, double ***conditionalF)
{

  /*Fills G for level g according to current configuration on levels 0...g-1*/

  int t,i,j,l,k,h;
  int nchildren,component,Father,Mother,ntensor,ntensor0,nparents,child,father,mother;
  pair *ctensor,**(ttensor);
  double *proba,sumproba;

  nchildren=C[g].n;

  if( (ctensor = (pair*) malloc(nchildren*sizeof(pair))) == NULL)
    alloc_error("find_configurations with 'ctensor'");

  for(i=0; i < nchildren; i++)
    {
      l=C[g].parent_at_level[i][0];
      if(l==-1)
	{
	  ctensor[i][0]=genes[C[g].individuals[i]][marker][0];
	  ctensor[i][1]=genes[C[g].individuals[i]][marker][1];
	}
      else
	{
	  t=C[g].parent_at_level[i][1];
	  k=C[l].p_comp[t];
	  h=C[l].pind_in_comp[t];
	  
	  ctensor[i][0] = G[l].comp[k].tensor[G[l].ordered[G[l+1].prev_config].component[k]][h][0];
	  ctensor[i][1] = G[l].comp[k].tensor[G[l].ordered[G[l+1].prev_config].component[k]][h][1];
	}
    }

  if( (G[g].nconfig = (int *) malloc(C[g].ncomponents*sizeof(int))) == NULL)
     alloc_error("find_configurations with 'G.nconfig'");

  if( (G[g].comp = (c_list *) malloc(C[g].ncomponents*sizeof(c_list))) == NULL)
     alloc_error("find_configurations with 'G.comp'");


  for(component=0; component < C[g].ncomponents; component++)
    {

      nchildren = C[g].components[component].nchildren;
      nparents = C[g].components[component].nparents;

      if( (ttensor = (pair**) malloc( sizeof(pair*))) == NULL)
	alloc_error("find_configurations with 'ttensor'");

      if( (ttensor[0] = (pair*) calloc(nparents, sizeof(pair))) == NULL)
	alloc_error("find_configurations with 'ttensor'");

     
      ntensor=1;

      for(j=0; j<C[g].components[component].nknowndata; j++)
	{
	  k=C[g].components[component].knowndata[j];
	  t=C[g].components[component].parents[k];

	  ttensor[0][k][0]=genes[t][marker][0];
	  ttensor[0][k][1]=genes[t][marker][1];
	}

      for(j=0; j<C[g].components[component].ncloned; j++)
	{
	  t=C[g].components[component].parents[C[g].components[component].cloned[j][0]];
	  l=C[g].components[component].cloned[j][1];
	  h=ordinal(t,C[l].parents,C[l].nparents);
	  k=C[l].p_comp[h];
	  t=C[g].components[component].cloned[j][0];

	  ttensor[0][t][0]=G[l].comp[k].tensor[G[l].ordered[G[l+1].prev_config].component[k]][C[l].pind_in_comp[h]][0];
	  ttensor[0][t][1]=G[l].comp[k].tensor[G[l].ordered[G[l+1].prev_config].component[k]][C[l].pind_in_comp[h]][1];

	}

      for(j=0;j<nchildren ;j++)
	{

	  child=C[g].components[component].children[j];
	  t=C[g].components[component].cind[j];

	  Father = fixed_parents[child][0];
	  father = ordinal(Father,C[g].components[component].parents,C[g].components[component].nparents);

	  Mother = fixed_parents[child][1];
	  mother = ordinal(Mother,C[g].components[component].parents,C[g].components[component].nparents);

	  ntensor0=ntensor;

	  for(i=0;i<ntensor0;i++)
	    ttensor=update_list(ctensor[t],father,Father,mother,Mother,ttensor,&i,&ntensor,&ntensor0, conditionalF,nparents,N_alleles[marker]);

	  for(i=0;i<ntensor;i++)
	    join_identical2(ttensor,&i,&ntensor,nparents);
	      
	}
    
      if(ntensor>0)
	{

	  if ( (proba = (double *) malloc(ntensor*sizeof(double))) == NULL)
	     alloc_error("find_configurations with 'proba'");

	  for(i=0;i<ntensor;i++)
	    {

	      proba[i]=1.0;
	      for(j=0;j<nparents;j++)
		{
		  k=C[g].components[component].parents[j];
		  proba[i] *= conditional_probability(k,ttensor[i][j],conditionalF,N_alleles[marker]);
		}
	    }

	  ntensor0=ntensor;
	  order_component(ttensor,proba,ntensor);

	  if(M_comp>0 && ntensor>=M_comp)
	    ntensor = M_comp;

	  G[g].nconfig[component]=ntensor;

	  if( ((G[g].comp[component].tensor) = (pair**) malloc(ntensor * sizeof(pair*))) == NULL)
	    alloc_error("find_configurations with 'G.comp.tensor'");

	  sumproba=0.0;

	  for(i=0;i<ntensor;i++)
	    {
	      if( (G[g].comp[component].tensor[i] = (pair*) malloc(nparents* sizeof(pair))) == NULL)
		alloc_error("find_configurations with 'G.comp.tensor'");

	      sumproba += proba[i];
	    }

	  if( ((G[g].comp[component].proba) = (double*) malloc(ntensor * sizeof(double))) == NULL)
	     alloc_error("find_configurations with 'G.comp.tensor'");

	  for(i=0;i<ntensor;i++)
	    {   
	      G[g].comp[component].proba[i]=proba[i]/sumproba;
	      
	      for(j=0;j<nparents;j++)
		{
		  G[g].comp[component].tensor[i][j][0]=ttensor[i][j][0];
		  G[g].comp[component].tensor[i][j][1]=ttensor[i][j][1];
		}
	      free(ttensor[i]);
	    }

	  for(i=ntensor;i<ntensor0;i++)
	    free(ttensor[i]);

	  free(proba);
	  free(ttensor);
	}
      else
	{   /* no consistent configurations found.*/
	
	  free(ttensor);
	  free(ctensor);

	  for(i=0;i<component;i++)
	    {
	      for(j=0;j<G[g].nconfig[i];j++)
		free(G[g].comp[i].tensor[j]);

	      free(G[g].comp[i].tensor);
	      free(G[g].comp[i].proba);

	    }

	  free(G[g].nconfig);
	  free(G[g].comp);

	  return component;
	
	}

    }

  G[g].n=0; //not yet defined, since may become too large
  G[g].log_n=0;

  for(component=0;component<C[g].ncomponents; component++)
    G[g].log_n += log(G[g].nconfig[component]);

  
  free(ctensor);

  return -1;
  
}
      


      
void free_glist(g_list *glist, int ncomponents)
{

  int j;
  unsigned i;

  for(i=0;(int) i<ncomponents;i++)
    {
      for(j=0;j<glist->nconfig[i];j++)
	free(glist->comp[i].tensor[j]);

      free(glist->comp[i].tensor);
      free(glist->comp[i].proba);
    }

  free(glist->comp);
  
  for(i=0;i<glist->n;i++)
    free(glist->ordered[i].component);

  if(glist->n > 0) free(glist->ordered);
  free(glist->nconfig);

  glist->n=0;
}

void join_identical2(pair **ttensor,int *i,int *ntensor,int nparents)
{

  /* Only those configurations in ttensor[] are saved that can not be reduced on
     each other.
  */

  int k,j;
  int identical=0;

  for(k=0;k<(*ntensor)&& !(identical);k++)
    {
      if(k != *i){

	identical=1;

	for(j=0;j<nparents && identical;j++)
	  identical = subtype(ttensor[*i][j],ttensor[k][j]);
      }
    }

  if(identical)
    delete_combination2(ttensor,i,ntensor);

}



int ordinal(int i, int *array, int n)
{
  register int *jp;
  int *last=array+n;

  for(jp=array;jp<last;++jp)
    if(*jp==i)
      return jp-array;

  return -1;
}




void sigproc()
{
  /* Ctrl-c prints used time and program exits.*/

  kello=clock()-kello;
  sek+=(double) kello / (double) CLOCKS_PER_SEC;
  printf("\nTime used %6.3f sec. \n",sek);
  //fprintf(times,"%6.3f\n",sek);
  exit(4);
}






int subtype(pair GT1, pair GT2){

  /*Returns 1 if GT1 is a special case of GT2, 0 otherwise.*/

  if(contains(GT2[0],GT1) && contains(GT2[1],GT1) && !( ( (GT2[0]==GT2[1]) && GT2[0]>0 ) && (GT1[0] != GT1[1])) )
    return 1;
  
  return 0;
}




pair** update_list(int child[2],int f,int F,int m,int M,pair **ttensor,int *i,int *ntensor,int *ntensor0, double ***conditionalF,int nparents,int n_alleles)
{

  int h,consistent,ncompatible=0,j,k;
  int father[2][2],mother[2][2];
  double pro;

  for(h=1-(child[0]==child[1]); h>=0; h--)
    {

       for(j=0;j<2;j++)
	{
	  father[ncompatible][j]=ttensor[*i][f][j];
	  mother[ncompatible][j]=ttensor[*i][m][j];
	}
       
       pro=1.0;

      consistent=compatible(child[1-h], father[ncompatible]);
      switch(consistent)
	{
	case 2:
	  pro = conditional_probability(F,father[ncompatible],conditionalF,n_alleles);

	case 1:
	  consistent=compatible(child[h], mother[ncompatible]);
	  
	  if(consistent==2)
	    pro *= conditional_probability(M,mother[ncompatible],conditionalF,n_alleles);

	case 0:
	  break;

	}

      if(consistent && pro>0.0)
	ncompatible++;
    }
  
  switch(ncompatible)
    {
    case 0:
      
      delete_combination(ttensor,i,ntensor,ntensor0);
      
      break;
      
    case 1:
      
      ttensor[*i][f][0]=father[0][0];
      ttensor[*i][f][1]=father[0][1];
      ttensor[*i][m][0]=mother[0][0];
      ttensor[*i][m][1]=mother[0][1];
      //join_identical(ttensor,i,ntensor,ntensor0,nparents);
		 
      break;

    case 2:
      
      (*ntensor)++;

      if ( (ttensor = (pair **) realloc(ttensor,(*ntensor)*sizeof(pair*))) == NULL)
	alloc_error("update_list with 'ttensor'");

      if ( (ttensor[(*ntensor)-1] = (pair *) malloc(nparents*sizeof(pair))) == NULL)
	alloc_error("update_list with 'ttensor'");
      
      for(k=0;k<nparents;k++)
	{
	  ttensor[(*ntensor)-1][k][0]=ttensor[*i][k][0];
	  ttensor[(*ntensor)-1][k][1]=ttensor[*i][k][1];
	}

      ttensor[*i][f][0]=father[0][0];
      ttensor[*i][f][1]=father[0][1];
      ttensor[*i][m][0]=mother[0][0];
      ttensor[*i][m][1]=mother[0][1];
      ttensor[(*ntensor)-1][f][0]=father[1][0];
      ttensor[(*ntensor)-1][f][1]=father[1][1];
      ttensor[(*ntensor)-1][m][0]=mother[1][0];
      ttensor[(*ntensor)-1][m][1]=mother[1][1];

      //join_identical(ttensor,i,ntensor,ntensor0,nparents);

      /*if(*ntensor > 0 )
	{
	  k=(*ntensor)-1;
	  join_identical(ttensor,&k,ntensor,ntensor0,nparents);
	  }*/
      break;
    }

  return ttensor;

}




void update_tgenes(pair *tgenes,int g, COMP *C,g_list *G,int step)
{

  /* tgenes (=temporary genes) is a structure that keeps track of the current configuration */

  int i,ind,c,p,a1,a2,j,consistent,config;

  // print_tgenes(tgenes,N_individuals,stdout);

  if(step==1)
    {
      for(i=0;i<C[g-1].nparents;i++)
	{
	  ind=C[g-1].parents[i];
	  c=C[g-1].p_comp[i];
	  p=C[g-1].pind_in_comp[i];
	  config=G[g-1].ordered[G[g].prev_config].component[c];
	  tgenes[ind][0]=G[g-1].comp[c].tensor[config][p][0];
	  tgenes[ind][1]=G[g-1].comp[c].tensor[config][p][1];
	}
    }
  if(step==-1) //Deletes the contribution of the parents at level g-1 from tgenes.
    {
      //G[g] has just been freed
      /*for(j=0;j<g+1;j++)
	{
	  for(i=0;i<C[j].nparents;i++)
	    {
	      ind=C[j].parents[i];
	      printf("%d:(%d,%d)\n",ind+1,tgenes[ind][0],tgenes[ind][1]=0);
	    }
	    }*/

      /*This could be made faster by additional structures that tells all the levels where an individual is a parent.
	Then it would suffice to delete the genotypes of people in C[g-1].parents from tgenes and check if some of those
	are parents also on some lower level and update tgenes accordingly.*/
	

      for(j=0;j<g+1;j++) 
	{
	  for(i=0;i<C[j].nparents;i++)
	    {
	      ind=C[j].parents[i];
	      tgenes[ind][0]=0;
	      tgenes[ind][1]=0;
	    }
	}

      for(j=g-1;j>=0;j--)
	{
	  for(i=0;i<C[j].nparents;i++)
	    {
	      ind=C[j].parents[i];
	      c=C[j].p_comp[i];
	      p=C[j].pind_in_comp[i];
	      config=G[j].ordered[G[j+1].prev_config].component[c];
	      a1=G[j].comp[c].tensor[config][p][0];
	      a2=G[j].comp[c].tensor[config][p][1];
	      consistent=compatible(a1, tgenes[ind]);
	      if(!consistent) 
		{
		  printf("Not consistent updating at level %d.\n",j);
		  exit(5);
		}
	      consistent=compatible(a2, tgenes[ind]);
	      if(!consistent) 
		{
		  printf("Not consistent updating at level %d.\n",j);
		  exit(5);
		}
	    } 
	}

      /*for(j=0;j<g+1;j++)
	{
	  for(i=0;i<C[j].nparents;i++)
	    {
	      ind=C[j].parents[i];
	      printf("%d:(%d,%d)\n",ind+1,tgenes[ind][0],tgenes[ind][1]);
	    }
	    }*/

    }
  //print_tgenes(tgenes,N_individuals,stdout);

}
  
