/* APE (Allelic Path Explorer) v.1.0
   searches for a consistent genotypic configuration 
   when the known pedigree structure 
   and genetic data for some individuals are given. 


   Copyright (C) Matti Pirinen
   
   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   version 2 as published by the Free Software Foundation.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program (gpl.txt); if not, write to the Free Software
   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, 
   MA  02110-1301, USA.

   www.gnu.org/licenses/gpl.html
*/


/*FILE: APE_print.c 
  Contains the output functions.
  See APE.h for the list of the functions.
*/




int check_compatibility(FILE *data,pair **g)
{

  /* 
     Checks whether the genes in g extend the 
     given datafile. 
  */

  int i,j,m=0,cons,a,ind,nzeros,ndata;
  pair *o;

  if((o=(pair*) calloc(N_markers,sizeof(pair)))==NULL)
    alloc_error("check_compatibility");

  fscanf(data,"%d",&ndata);
  if(ndata < N_knowndata)
    {
      fprintf(stderr,"ERROR: N_knowndata is larger than the one given in the datafile!\n");
      return 0;
    }

  cons=1;
  for(i=0;i<ndata && cons;i++)
    {
      fscanf(data,"%d",&ind);
      
      for(j=0;j<2;j++)
	for(m=0;m<N_markers;m++)
	  fscanf(data,"%d",&o[m][j]);

      for(m=0;m<N_markers && cons;m++)
	{
	  nzeros = (o[m][0]==0) + (o[m][1]==0);

	  switch(nzeros)
	    {
	    case 0:
	      if(!(((o[m][0]==g[ind-1][m][0]) && (o[m][1]==g[ind-1][m][1])) ||
		 ((o[m][0]==g[ind-1][m][1]) && (o[m][1]==g[ind-1][m][0]))))
		cons=0;
	      break;
	    case 1:
	      a=(o[m][0]>0) ? o[m][0] : o[m][1];
	      if(!((a==g[ind-1][m][0]) || (a==g[ind-1][m][1])))
		cons=0;
	      break;
	    case 2:
	      break;
	    }
	}
    }



  if(!cons)
    {
      printf("ERROR: individual %d marker %d not consistent with the data!.\n",ind,m);
      printf("Data: (%d,%d) but APE: (%d,%d)\n",o[m-1][0],o[m-1][1],g[ind-1][m-1][0],g[ind-1][m-1][1]);
      free(o);
      return 0;
    }
  printf("*** Data are consistent with the datafile! *** \n\n");
  free(o);
  return 1;
}




int check_consistency(pair **genes, int **fixed_parents)
{
  /*
    Checks that the data in 'genes' are Mendelian consistent i.e.
    at each marker each individual have inherited exactly one allele
    from his father and the other one from his mother.
  */

  int parents[4];
  int i,m,consistent;

  for(i=0;i<N_individuals;i++)
    {
      if(fixed_parents[i][0]>-1 && fixed_parents[i][1]>-1)
	{
	  for(m=0;m<N_markers;m++)
	    {
	      parents[0]=genes[fixed_parents[i][0]][m][0];
	      parents[1]=genes[fixed_parents[i][0]][m][1];
	      parents[2]=genes[fixed_parents[i][1]][m][0];
	      parents[3]=genes[fixed_parents[i][1]][m][1];
	      
	      consistent=( ( (genes[i][m][0]==parents[0] || genes[i][m][0]==parents[1] || genes[i][m][0]==0) &&
			    (genes[i][m][1]==parents[2] || genes[i][m][1]==parents[3] || genes[i][m][1]==0) ) 
			   ||
			   ( (genes[i][m][0]==parents[2] || genes[i][m][0]==parents[3] || genes[i][m][0]==0) &&
			   ( genes[i][m][1]==parents[0] || genes[i][m][1]==parents[1] || genes[i][m][1]==0) ) );
	      if(!consistent)
		{
		  printf("PROBLEMS WITH INDIVIDUAL %d: (%d,%d) at marker %d. \n",i+1,genes[i][m][0],genes[i][m][1],m+1);
		  printf("Father %d: (%d,%d)",fixed_parents[i][0],parents[0],parents[1]);
		  printf("and mother %d: (%d,%d). \nPedigree is INCONSISTENT!\n",fixed_parents[i][1],parents[2],parents[3]);
		  return 0;
		}
	    }
	}
    }

  printf("\n***** Data are Mendelian consistent! *****\n");
  return 1;
}



void draw_pedigree(char *pedfile)
{
  /*
    Requires pfiddler & list2pedc utility.
  */

  int i,ind,f,m,sex,n;
  FILE *ped=fopen(pedfile,"r");

  FILE *out=fopen(PEDIGREE_LIST_OUTPUT,"w");

  fscanf(ped,"%d\n",&n);

  for(i=0; i< n; i++)
    {
      fscanf(ped, "%d %d %d %d \n", &ind, &f, &m, &sex);
      fprintf(out,"%d %d %d %d \n",ind, f, m, sex);
    }

  fclose(out);
  fclose(ped);

  system("list2pedc -s pedlist -o pedlist.pedc");
  system("pfiddler pedlist.pedc &");


}



void print_combinations(config *ordered, int n, int ncomponents, int g,FILE *fp)
{

  int i,j;

  fprintf(fp,"Level %d:\n********\n\n",g);

  for(i=0;i<n;i++)
    {
      for(j=0; j< ncomponents;j++)
	fprintf(fp,"%d ",ordered[i].component[j]);
      
      fprintf(fp,"\n\n");
    }
}


void print_comments(int success)
{
  /* values of 'success'
      0 : found a configuration
      1 : didn't find, returned to the 0 level
      2 : didn't find, maximum number of backward jumps encountered
  */

  printf("\nOrdering parameters: kappa=%6.3f, w=%6.3f.\n",kappa,diff_exp);
  printf("Reduction parameters: M_comp=%d, M_lev=%d, M_clu=%d.\n",M_comp,M_lev,M_clu);

  if((success==1 && (M_comp > 0 || M_lev > 0 || M_clu > 0)) || success == 2 )
    {
      printf("\nThere may still be unchecked configurations!\n"); 
    }

  kello=clock()-kello;
  sek+=(double) kello / (double)CLOCKS_PER_SEC;
  printf("\nTime used %6.3f sec. \n",sek);
  //fprintf(times,"%10.5g",sek);

  /*if(success)
    fprintf(times," %d\n",2);
  else
  fprintf(times," %d\n",0);*/
}


void print_components(COMP *C, FILE *fp)
{

  int i,g,j,t,l;

  fprintf(fp,"\n**** Printing components ****\n");

  for(g=0;g<N_levels-1;g++)
    {
      fprintf(fp,"\n\n Level %d, %d components and %d individuals.\n\n ************************************** ",g,C[g].ncomponents,C[g].n);
      for(i=0;i<C[g].ncomponents;i++)
	{
	  fprintf(fp,"Component %d\n\n %d parents:\n", i, C[g].components[i].nparents);
	  for(j=0;j<C[g].components[i].nparents;j++)
	    fprintf(fp,"%d ",C[g].components[i].parents[j]+1);
	  fprintf(fp,"\n %d parents with known genotype:\n", C[g].components[i].nknowndata);
	  for(j=0;j<C[g].components[i].nknowndata;j++)
	    fprintf(fp,"%d ",C[g].components[i].parents[C[g].components[i].knowndata[j]]+1);
	  fprintf(fp,"\n %d parents with clones at lower levels:\n", C[g].components[i].ncloned);
	  for(j=0;j<C[g].components[i].ncloned;j++)
	    fprintf(fp,"(%d,%d) ",C[g].components[i].parents[C[g].components[i].cloned[j][0]]+1,C[g].components[i].cloned[j][1]);
	  fprintf(fp,"\n\n %d children:\n", C[g].components[i].nchildren);
	  for(j=0;j<C[g].components[i].nchildren;j++)
	    fprintf(fp,"%d ",C[g].components[i].children[j]+1);
	  fprintf(fp,"\n who have children at level:\n");
	  for(j=0;j<C[g].components[i].nchildren;j++)
	    {
	      t=ordinal(C[g].components[i].children[j],C[g].individuals,C[g].n);
	      l=C[g].parent_at_level[t][0];
	      if(l>-1)
		fprintf(fp,"(%d,%d) ",l,C[C[g].parent_at_level[t][0]].parents[C[g].parent_at_level[t][1]]+1);
	      else
		fprintf(fp,"(%d) ",l);
		
	    }
	  fprintf(fp,"\n\n");
  
	}
      fprintf(fp,"\n");
    }

  fprintf(fp,"\n\nLevel %d, %d components and %d individuals.\n\n ****************************************** ",g,C[g].ncomponents,C[g].n);
  for(j=0;j<C[g].n;j++)
    fprintf(fp,"%d ",C[g].individuals[j]+1);
  fprintf(fp,"\n\n**** END OF THE COMPONENTS LIST ****\n\n");
}





void print_components_info(COMP *C, FILE *fp)
{

  int g,t,l,max,np,nc;
  int nuclear=0;
  int components=0;
  
  np=nc=0;
  fprintf(fp,"\n**** Printing info about components ****\n");
  
  for(g=0;g<N_levels-1;g++)
    {
      fprintf(fp,"\n Level %d, %d components and %d individuals.\n",g,C[g].ncomponents,C[g].n);
      fprintf(fp,"On average each component contains %f parents and %f children.\n",1.0*C[g].nparents/C[g].ncomponents,1.0*C[g].n/C[g].ncomponents);
      np += C[g].nparents;
      nc += C[g].n;
      max=0;
      for(t=0;t<C[g].ncomponents;t++){
	components++;
	if(C[g].components[t].nparents==2) nuclear++;
	l=C[g].components[t].nparents+C[g].components[t].nchildren;
	max= (l < max) ? max : l; 
	fprintf(fp,"%d + %d = %d \n",C[g].components[t].nparents,C[g].components[t].nchildren,C[g].components[t].nparents+C[g].components[t].nchildren); 
      }
      
      printf("%d (%3.1f,%3.1f) ",max,1.0*C[g].nparents/C[g].ncomponents,1.0*C[g].n/C[g].ncomponents);
    }
  
  
  printf("\nNuclear families %d/%d=%f. Average component size %3.1f p and %3.1f c.\n Founders: %d\n",nuclear,components,1.0*nuclear/components,1.0*np/components,1.0*nc/components,C[N_levels-1].n);
  

}



void print_conditionalF(double ***conditionalF, COMP *C, FILE *fp, int n_alleles)
{

  int i,g,a,b,j;

  fprintf(fp,"\n\n **** PRINTING THE LIST OF THE CONDITIONAL PROBABILITIES ****\n\n");
  for(g=0; g<N_levels; g++)
    {
      fprintf(fp,"\nLevel %d\n***************\n",g);

      for(j=0;j<C[g].n;j++)
	{
	  i=C[g].individuals[j];
	  fprintf(fp,"individual %d\n",i+1);

	  for(a=0;a<n_alleles; a++)
	    {
	      for(b=0;b<n_alleles; b++)
		fprintf(fp,"{%d,%d} = %f ",a+1,b+1,conditionalF[i][a][b]);
	      fprintf(fp,"\n");
	    }
	}
    }
  fprintf(fp,"\n**** END OF THE LIST OF THE CONDITIONAL PROBABILITIES ****\n\n");

}

void print_connected_components(int **parents)
{
  /*
    Checks whether the pedigree is connected or can be divided into several components.
    If it is not connected, the division is printed into a file specified by the macro 
    CONNECTED_COMPONENT_OUTPUT
  */


  int i,j;
  int *comp;
  int ncomp=1;
  int *colors;
  FILE *fp;

  if( (comp=(int*) calloc(N_individuals,sizeof(int)))==NULL)
    alloc_error("print_connected_components");

  for(i=0;i<N_individuals;i++)
    {
      /*for(j=0;j<N_individuals;j++)
	printf("%d %d\n",j+1,comp[j]);*/
      if(!comp[i])
	{
	  if(ncomp==set_component(i,ncomp,parents,comp))
	    ncomp++;
	}
    }
  
  if( (colors=(int*) malloc(ncomp*sizeof(int)))==NULL)
    alloc_error("print_connected_components");
  
  ncomp=0;
  for(i=0;i<N_individuals;i++)
    if(ordinal(comp[i],colors,ncomp)==-1)
      colors[ncomp++]=comp[i];

  if(ncomp>1)
    {
      printf("\nNOTE: The pedigree has %d connected components.\n",ncomp);
      printf("See the file %s for the appropriate division.\n",CONNECTED_COMPONENTS_OUTPUT);
      fp=fopen(CONNECTED_COMPONENTS_OUTPUT,"w");
      for(j=0;j<ncomp;j++)
	{
	  fprintf(fp,"Component %d:\n",j+1);
	  for(i=0;i<N_individuals;i++)
	    if(comp[i]==colors[j])
	      fprintf(fp,"%d\n",i+1);
	  fprintf(fp,"\n");
	}
      fclose(fp);
    }
  else
    printf("\nPedigree is connected.\n");

  free(comp);
  free(colors);

}


void print_defined_genes(pair **genes,COMP *C)
{
  int i,l,defined,ind;

  printf("Defined alleles for each level (count and ratio).\n");
  for(l=0;l<N_levels;l++)
    {
      defined = 0;
      for(i=0;i<C[l].n;i++)
	{
	  ind = C[l].individuals[i];
	  defined += (genes[ind][0][0]>0) + (genes[ind][0][1]>0);
	} 
      printf("%3d: %6d, %4.2f\n",l,defined,defined/2.0/C[l].n);
    }
}

void print_generation(g_list *G,COMP *C, int g,FILE *fp)
{

  int n,i,j,k,c;
	      
  fprintf(fp,"\n******Level %d ****** \n",g);
  if(g>0) 
    {
      fprintf(fp,"Previous configuration: ");
      for(i=0;i<C[g-1].ncomponents;i++)
	fprintf(fp," %d ",G[g-1].ordered[G[g].prev_config].component[i]);
      fprintf(fp,"\n");
    }
  if(G[g].nconfig[0]==0)
    fprintf(fp,"NULL \n");
  else
    {
      for(c=0;c<C[g].ncomponents;c++)
	{
	  n=C[g].components[c].nparents;
	  fprintf(fp,"Component %d, %d x %d pairs.\n",c,G[g].nconfig[c],n);
	      
	  for(j=0;j<G[g].nconfig[c];j++)
	    {
	      fprintf(fp,"%f ",G[g].comp[c].proba[j]);
	      fprintf(fp,"(%d,%d)",G[g].comp[c].tensor[j][0][0],G[g].comp[c].tensor[j][0][1] );
	      for(k=1;k<n;k++)
		fprintf(fp,"x(%d,%d)",G[g].comp[c].tensor[j][k][0],G[g].comp[c].tensor[j][k][1]);
	      fprintf(fp,"\n"); 
	    }

	  fprintf(fp,"\n");
	}
    }
}


void print_genotypes(g_list *G, COMP *C, int g, FILE *fp)
{

  /*Prints current configuration at lower levels.*/

  int i,l,g1=0,g2=0,ind,c,p;

  for(l=0;l<=g;l++)
    {
      fprintf(fp,"\n Level %d\n",l);

      for(i=0;i<C[l].nparents;i++)
	{
	  ind=C[l].parents[i];
	  c=C[l].p_comp[i];
	  p=C[l].pind_in_comp[i];
	  if(l<g)
	    {
	      g1=G[l].comp[c].tensor[G[l].ordered[G[l+1].prev_config].component[c]][p][0];
	      g2=G[l].comp[c].tensor[G[l].ordered[G[l+1].prev_config].component[c]][p][1];
	    }
	  else if(l<N_levels-2)
	    {
	      g1=G[l].comp[c].tensor[G[l].ordered[0].component[c]][p][0];
	      g2=G[l].comp[c].tensor[G[l].ordered[0].component[c]][p][1];
	    }
	  fprintf(fp,"%d: (%d,%d)\n",ind+1,g1,g2);
	}
    }
}




void print_inbred(FILE *fp, double **kinship)
{
 /*
    Note that kinship is a matrix that contains all pairwise
    kinship coefficients. See count_kinship() for making it.
  */

  int i,inbred=0;

  fprintf(fp,"Printing inbred individuals (index, kinship with himself/herself, inbred coeff.)\n");
  for(i=0;i<N_individuals;i++)
    if(kinship[i][i]>0.5)
      {
	fprintf(fp,"%d %6.4f %6.4f\n",i+1,kinship[i][i],2*(kinship[i][i]-0.5));
	inbred++;
      }
  printf("Inbred individuals %d/%d = %6.4f.\n",inbred,N_individuals,1.0*inbred/N_individuals);
  fprintf(fp,"%d %f\n",inbred,1.0*inbred/N_individuals);
}


void print_inconsistencies(FILE *out,UL **inconsistencies,COMP *C,UL no_cons,double frac)
{

  int l,c;
  UL max;
  int lmax=-1,cmax=-1;

  printf("\nPrinting highest inconsistency rates (limit %4.2f).\n",frac);

  for(;;)
    {
      max=0;
      for(l=0;l<N_levels;l++)
	{
	  for(c=0;c<C[l].ncomponents;c++)
	    {
	      if(inconsistencies[l][c]>max)
		{
		  max=inconsistencies[l][c];
		  lmax=l;
		  cmax=c;
		}
	    }
	}
      if(1.0*max/no_cons > frac)
	{
	  printf("level=%d,component=%d: %f\n",lmax,cmax,1.0*max/no_cons);
	  inconsistencies[lmax][cmax]=0;
	  fprintf(out,"%d %d\n",lmax,cmax);
	}
      else
	break;
    }
}


void print_kinship_matrix(double **kinship,FILE *fp)
{
  /*
    Note that kinship should contain all pairwise
    kinship coefficients, not just those
    for parents belonging to the same level.
    See count_kinship() for possible way to make the matrix.
  */

  int i,j;

  fprintf(fp,"\n \n*** Printing kinship coefficients. ***\n\n");

  for(i=0;i<N_individuals;i++)
    {
      fprintf(fp,"\n %d:",i+1);
      for(j=0;j<N_individuals;j++)
	{
	  if(j%10==9) fprintf(fp,"\n %d:",i+1);
	  fprintf(fp,"(%d)=%f  ",j+1,kinship[i][j]);
	}
    }
  fprintf(fp,"\n");

}

void print_kinship(FILE *fp,COMP *C,double ***kinship)
{

  int i,j,g;

  fprintf(fp,"\n \n*** Printing kinship coefficients. ***\n\n");

  for(g=0;g<N_levels-1;g++)
    {
      fprintf(fp,"\nParents at level %d\n",g);
      for(i=0;i<C[g].nparents;i++)
	{
	  fprintf(fp,"\n %d:",C[g].parents[i]);
	  for(j=0;j<C[g].nparents;j++)
	    {
	      if(j%10==9) 
		fprintf(fp,"\n %d:",C[g].parents[i]);
	      fprintf(fp,"(%d)=%f  ",C[g].parents[j],kinship[g][i][j]);
	    }
	}
    }
  fprintf(fp,"\n");

}


void print_locus(char *pedfile, pair **genes,int marker,FILE *out)
{
  int i,ind,f,m,sex,n;
  FILE *ped=fopen(pedfile,"r");
 
  fscanf(ped,"%d\n",&n);

  for(i=0; i< n && !(feof(ped)); i++)
    {

      fscanf(ped, "%d %d %d %d \n", &ind, &f, &m, &sex);
      fprintf(out,"%d %d %d %d \n",ind, f, m, sex);

      fprintf(out,"%d ",genes[ind-1][marker][0]);
      fprintf(out,"\n");
      fprintf(out,"%d ",genes[ind-1][marker][1]);
      fprintf(out,"\n\n");	
    }
  fclose(ped);
}

void print_offspring(CHILDREN *offspring,FILE *out)
{
  int i,j;

  fprintf(out,"List of Children\n\n");

  for(i=0;i<N_individuals;i++)
    {
      fprintf(out,"%d has %d children:\n",i+1,offspring[i].nchildren);

      for(j=0;j<offspring[i].nchildren;j++)
	  fprintf(out,"%d ",offspring[i].children[j]+1);

      fprintf(out,"\n\n");
    }
}


void print_parental_info(FILE *out, PARENTAL* parental)
{
  int i,l;

  for(i=0;i<N_individuals;i++)
    {
      fprintf(out,"%d (%d):",i+1,parental[i].nlevels);
      for(l=0;l<parental[i].nlevels;l++)
	fprintf(out,"%d ",parental[i].levels[l]);
      fprintf(out,"\n");
    }

}

void print_pedigree(char *pedfile, pair **genes,int last_marker,FILE *out)
{

  /*Does simple output:
    only alleles that APE has transmitted up to founders are shown.
    There is no phase information in this output.*/

  int marker,i,ind,f,m,sex,n;
  FILE *ped=fopen(pedfile,"r");
 
  fscanf(ped,"%d\n",&n);

  for(i=0; i< n && !(feof(ped)); i++)
    {

      fscanf(ped, "%d %d %d %d \n", &ind, &f, &m, &sex);
      fprintf(out,"%d %d %d %d \n",ind, f, m, sex);

      for(marker=0; marker<=last_marker; marker++)
	fprintf(out,"%d ",genes[ind-1][marker][0]);
      fprintf(out,"\n");
      for(marker=0;marker<=last_marker; marker++)
	fprintf(out,"%d ",genes[ind-1][marker][1]);
      fprintf(out,"\n\n");	
    }
  fclose(ped);
}






void print_pedigree_info(FILE *fp,COMP *C,int *sexes)
{
  /*Prints info about the ratio of males and females.*/

  int i,male,female;
  int g=-(N_levels-1);

  fprintf(fp,"%d %d\n",g,C[-g].n);
  for(g=N_levels-2;g>=0;g--)
    {
      male=female=0;
      for(i=0;i<C[g].nparents;i++)
	if(sexes[C[g].parents[i]]==1)
	  ++male;
	else
	  ++female;

      fprintf(fp,"%d %d %f %f ",-g,C[g].n,1.0*C[g].nparents/C[g].ncomponents,1.0*C[g].n/C[g].ncomponents);
      fprintf(fp,"%f %f\n",1.0*male/C[g].ncomponents,1.0*female/C[g].ncomponents);
    }

}




void print_phase(int **phase, FILE *fout)
{

  int i,m;

  fprintf(fout,"\n\n Printing phase vectors \n\n");

  for(i=0;i< N_individuals; i++)
    {
      fprintf(fout,"%d: ",i+1);

      for(m=0;m<N_markers;m++)
	fprintf(fout,"%d ",phase[i][m]);

      fprintf(fout,"\n");
    }
}


void print_tgenes(pair *tgenes, FILE *fp)
{
  int i;

  for(i=0;i<N_individuals;i++)
    {
      fprintf(fp,"%d:(%d,%d)\n",i+1,tgenes[i][0],tgenes[i][1]);
    }

}



void print_total(char *pedfile,pair **genes,pair **origins, int *levels,FILE *out)
{
  /*Total output.*/

  int marker,i,ind,f,m,sex,n;
  FILE *ped=fopen(pedfile,"r");
 
  fscanf(ped,"%d\n",&n);

  for(i=0; i< n && !(feof(ped)); i++)
    {

      fscanf(ped, "%d %d %d %d \n", &ind, &f, &m, &sex);
      fprintf(out,"%d %d %d %d %d\n",ind, f, m, sex, levels[ind-1]);

      for(marker=0;marker<N_markers;marker++)
	fprintf(out,"%d ",genes[ind-1][marker][0]);
      fprintf(out,"\n");
      for(marker=0;marker<N_markers;marker++)
	fprintf(out,"%d ",origins[i][marker][0]);
      fprintf(out,"\n");
      for(marker=0;marker<N_markers;marker++)
	fprintf(out,"%d ",genes[ind-1][marker][1]);
      fprintf(out,"\n");
      for(marker=0;marker<N_markers;marker++)
	fprintf(out,"%d ",origins[i][marker][1]);
      fprintf(out,"\n\n");	
    }
  fclose(ped);
}

int set_component(int ind,int color,int **parents,int *comp)
{

  /*A recursively called function for print_connected_component
    Returns the value "color" with which it has colored
    ind, his/her parents and the component above them.
  */

  int i;
  int f=parents[ind][0];
  int m=parents[ind][1];
 
  comp[ind]=color;

  if(f>-1 && m>-1)
    {
      if(!comp[f])
	color=set_component(f,color,parents,comp);
      else
	{
	  for(i=0;i<N_individuals;i++)
	    if(comp[i]==color)
	      comp[i]=comp[f];
	  color=comp[f];
	}
      if(!comp[m])
	color=set_component(m,color,parents,comp);
      else
	{
	  for(i=0;i<N_individuals;i++)
	    if(comp[i]==color)
	      comp[i]=comp[m];
	  color=comp[m];
	}
    }
  return color;
}

