/* APE (Allelic Path Explorer) v.1.0
   searches for a consistent genotypic configuration 
   when the known pedigree structure 
   and genetic data for some individuals are given. 


   Copyright (C) Matti Pirinen
   
   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   version 2 as published by the Free Software Foundation.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program (gpl.txt); if not, write to the Free Software
   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, 
   MA  02110-1301, USA.

   www.gnu.org/licenses/gpl.html
*/

/*FILE: APE_extract.c
  A program that extracts components specified in the file 
  "extracted_component"
  See the end of the file for details.
*/

#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include <signal.h>
#include <limits.h>
#include "APE.h"


int **extract_component(int g, int c0, int *nind, COMP *C);
int **extract_component2(int g, int c0,int *nind, COMP *C,CHILDREN *offspring, int *levels);
int **extract(pair *extracted, int n_ex, int max_g, int *nind, COMP *C, CHILDREN *offspring, int *levels);
void undefine_rest(int *nind, int **ind, pair **genes, int marker, int *levels, int g);
int find_extracted_configurations(g_list *G, COMP *C, int g, int **fixed_parents, pair** genes, int marker, double ***conditionalF,int **ind, int *nind,int ex_level,int *levels);
void print_extracted(FILE* out,COMP *C,pair **genes, pair *extracted, int n_ext,int markers);
void print_extracted_component(FILE* out,COMP *C,pair **genes, int g,int c0,int markers);
void write_new_datafile(FILE *data, pair **genes, COMP* C, int *knowndata, pair *extracted, int n_ext);
int write_new_datafile2(FILE *data, pair **genes, int *knowndata, int max_g, int **ind, int* nind, pair *extracted, int n_ext, COMP *C);
void read_component(FILE *compfile, int *l, int *c,COMP *C);
pair *read_extracted(FILE *ext,int *n_extracted,int *max_ext_level, COMP *C);
void print_suggestions(FILE* out,COMP *C,pair *extracted, int n_ext);



int main(int argc, char* argv[])
{

  char pedfile[40],datafile[40],afreqfile[40];
  int i,g,marker,j;
  int *knowndata;
  int **fixed_parents;
  COMP *C;
  g_list *G;
  pair **genes, *tgenes;
  double ***conditionalF,***kinship,**afreq,*ratio_known;
  int *levels;
  int allowed_jumps; 
  CHILDREN *offspring;
  PARENTAL *parental;
  int n_ext,max_ext_level;
  int **ext_ind, *n_ext_ind;
  UL **inconsistencies;
  pair *extracted;
  clock_t kello2;


  kello=clock();
  signal(SIGINT, sigproc);
  srand( (unsigned int) time(NULL) );


  /********************** Initializing the known pedigree/data structure. **************************/

  printf("\n*** APE_extract ***\n");
  if(argc>1)
    {
      FILE *args;
    
      if((args= fopen(argv[1],"r"))==NULL)
	{fprintf(stderr, "\nError: Cannot open parameters file \"%s\"!\nCorrect usage:\"APE parmeters_file\"\nSee README for details about parameters.\n\n",argv[1]);
	 exit(3);}
    
      read_parameters(args,pedfile,datafile,afreqfile,&N_markers, &M_comp, &M_lev, &kappa, &diff_exp, &M_clu, &allowed_jumps);
      fclose(args);
    }
  else
    { 
      fprintf(stderr, "\nError: No parameters file specified!\nCorrect usage:\"APE parmeters_file\"\nSee README for details.\n\n");
      exit(3);
    }

  //times=fopen("times","a"); // This file is for test runs only.
  //fprintf(times,"E");

 /* reads the known pedigree from file */
  { 
    FILE *ped;
    if((ped= fopen(pedfile,"r"))==NULL)
      {
	fprintf(stderr, "\nError: Cannot open pedigree file \"%s\"!\nSee README for details about parameters.\n\n",pedfile);
	exit(3);
      }
    fixed_parents = readPedigree(ped,&N_individuals);
    fclose(ped); 
  }  

  printf("individuals %d\n",N_individuals);
  
  offspring = list_children(fixed_parents);
  //print_offspring(offspring,stdout);

  if( (genes = (pair **) malloc(N_individuals * sizeof(pair*)) ) == NULL)
    alloc_error("main with genes");
  
  for(i=0; i<N_individuals ;i++)
    if( (genes[i] = (pair *) calloc(N_markers, sizeof(pair)) ) == NULL)
      alloc_error("main with genes");

 if( (tgenes = (pair *) malloc(N_individuals * sizeof(pair)) ) == NULL)
   alloc_error("main with tgenes");

  { 
    FILE *data;
    if((data= fopen(datafile,"r"))==NULL)
      {
	fprintf(stderr, "\nError: Cannot open data file \"%s\"!\nSee README for details about parameters.\n\n",datafile); 
	exit(3);
      }
    knowndata = readData(data,genes,&N_knowndata);  
    fclose(data); 
    infer_from_homozygosity(fixed_parents,genes); 

    if((data= fopen(afreqfile,"r"))==NULL)
      {
	fprintf(stderr, "\nError: Cannot open allele frequency file \"%s\"!\n",afreqfile);
	fprintf(stderr, "See README for details about parameters.\n\n"); 
	exit(3);
      }
    afreq = readAfreq(data);
    fclose(data);
  }

  if( (levels = (int*) malloc(N_individuals*sizeof(int))) == NULL)
    alloc_error("main with levels");
  
  for(i=0;i<N_individuals;i++)
    levels[i]=-1;

  N_levels=0;
  
  for(i=0;i<N_individuals;i++)
   {
     count_level(i,fixed_parents,levels);
     N_levels=(levels[i]+1>N_levels) ? levels[i]+1 : N_levels;
   }
  
  if ((C = (COMP *) malloc(N_levels * sizeof(COMP)) ) == NULL)
    alloc_error("main with C");

  make_components(C,fixed_parents,knowndata,levels);
  parental = find_parental_info(C);
  kinship=count_kinship2(C,fixed_parents,parental);
  find_component_ordering(C,kinship);

  if( (G = (g_list *) malloc((N_levels-1) * sizeof(g_list)) ) == NULL)
    alloc_error("main with G");

  if( (inconsistencies=(UL**) malloc(N_levels*sizeof(UL*))) == NULL)
    alloc_error("main with inconsistencies");
  for(i=0;i<N_levels;i++)
    if( (inconsistencies[i] = (UL*) calloc(C[i].ncomponents,sizeof(UL))) == NULL)
      alloc_error("main with inconsistencies");
    
  {
    FILE *ext;
    if( (ext = fopen(EXTRACTIONS_FILE,"r")) ==NULL)
      {
	fprintf(stderr, "\nError: Cannot open %s file!\nSee README for details.\n\n",EXTRACTIONS_FILE);
	exit(3);
      }
    extracted=read_extracted(ext,&n_ext,&max_ext_level,C);
    fclose(ext);
  }
  
  if( (n_ext_ind = (int *) calloc(max_ext_level+2,sizeof(int))) == NULL)
    alloc_error("main with n_ext_ind");

  ext_ind=extract(extracted, n_ext, max_ext_level, n_ext_ind,C,offspring,levels);


 /************************ Checking all the possible configurations. ************************/


  for(marker=0;marker<N_markers;marker++)
    {

      int new, inconsistent_comp;
      unsigned long jump_back=0;

      conditionalF = transmission_probabilities( genes, fixed_parents, C, marker, N_alleles[marker], afreq[marker]);
      undefine_rest(n_ext_ind, ext_ind,genes,marker,levels,max_ext_level);
      check_consistency_of_components(C,N_levels,fixed_parents,genes,marker,conditionalF);
      ratio_known=count_known_data(genes,marker,C);

      {
	if(marker>0)
	  {
	    G[N_levels-2].n=0; // Since this was not ordered during the previous round.
	    for(i=0;i<N_levels-1;i++)
	      free_glist(&G[i],C[i].ncomponents);
	  }
	
	for(i=0; i<N_levels-1; i++)
	  G[i].prev_config=0;
	
	for(i=0; i<N_individuals;i++)
	  {
	    tgenes[i][0]=genes[i][marker][0];
	    tgenes[i][1]=genes[i][marker][1];
	  }
	
	g=0;
	
	do
	  {
	    
	    inconsistent_comp=find_extracted_configurations(G, C, g, fixed_parents, genes, marker, conditionalF,ext_ind,n_ext_ind,max_ext_level,levels);
	    
	    if(inconsistent_comp == -1)
	      {
		printf("%d: ",g);
		if(g>0)
		  update_tgenes(tgenes,g,C,G,1); //Updates temporary configuration to the level g i.e. current configurations at levels 0,...,g-1 are fixed. 
		if(g<N_levels-2)
		  {
		    //constrain_components(G,C,g);
		    //order_generation2(G,C,g,kinship,tgenes,fixed_parents,offspring,levels);
		    order_generation(G,C,g,kinship,tgenes,ratio_known);
		    //print_combinations(G[g].ordered,G[g].n,C[g].ncomponents,g,outp);
		  }
		else
		  build_pedigree(genes,G,C,marker);
		
		//print_generation(G,C,g,outp);
		//print_genotypes(G,C,g,outp);
		g++;
	      }
	    else
	      {
		inconsistencies[g][inconsistent_comp]++;
		no_cons++;
		
		if(! ( new= (++G[g].prev_config != G[g-1].n) ) )
		  {
		    G[g].prev_config=0;
		    kello2=clock();
		    if( ((double) (kello2-kello) / (double) CLOCKS_PER_SEC)>100)
		      {
			sek+=(double)(kello2-kello)/(double) CLOCKS_PER_SEC;
			kello=kello2;
		      }
		  }

		while(!new && g > 0)
		  {
		    jump_back++;
		    if( jump_back >= (unsigned) allowed_jumps && allowed_jumps > 0)
		      {
			printf("\nProgram aborted since %ld jumps were encountered.\n",jump_back);
			{  
			  FILE *out=fopen(EXTRACTIONS_FILE,"w");
			  print_suggestions(stdout,C,extracted,n_ext);
			  print_inconsistencies(out,inconsistencies,C,no_cons,inconsistency_frac);
			  fclose(out);
			}
			print_comments(2);
			exit(2);
		      }
		    g--;
		    free_glist(&G[g],C[g].ncomponents);
		    G[g].prev_config++;
		    if(g>0 && (unsigned) G[g].prev_config==G[g-1].n)
		      G[g].prev_config=0;
		    
		    
		    update_tgenes(tgenes,g,C,G,-1);
		    
		    new=(G[g].prev_config>0);
		  }
		
	      }
	    
	  }while(g>0 && g < N_levels-1);
	
	if(g==0)
	  {
	    printf("Inconsistent data at marker %d. Program aborted!\n",marker+1);
	    printf("unsuccessful configurations=%ld; backward jumps=%ld\n", no_cons, jump_back);
	    {  
	      FILE *out=fopen(EXTRACTIONS_FILE,"w");	    
	      print_suggestions(stdout,C,extracted,n_ext);
	      print_inconsistencies(out,inconsistencies,C,no_cons,inconsistency_frac);
	      fclose(out);
	    }
	    print_comments(1);
	    exit(2);
	  }
      }
      printf("Consistent configuration found for marker %d. \n",marker+1);
      printf("unsuccessful configurations=%ld; backward jumps=%ld\n", no_cons, jump_back);
      
      for(i=0;i<N_individuals;i++)
	{
	  for(j=0;j<N_alleles[marker];j++)
	    free(conditionalF[i][j]);
	  free(conditionalF[i]);
	}
      free(conditionalF);

    }	     /* for(marker=0;marker<N_markers)*/


  //print_extracted(stdout,C,genes,extracted,n_ext,N_markers);  
  
  {
    FILE *data = fopen(datafile,"r");
    //write_new_datafile(data,genes,C, knowndata,extracted,n_ext);
    write_new_datafile2(data,genes,knowndata,max_ext_level,ext_ind,n_ext_ind,extracted,n_ext,C); 
    //function closes "data" and writes to the file EXTRACTED_DATA_FILE
  }
  
  print_comments(0);

  return 0;
}


#include "APE_initialize.c"
#include "APE_general.c"
#include "APE_order.c"
#include "APE_print.c"



/*** EXTRACTION FUNCTIONS ***/




int **extract_component(int g, int c0, int *nind, COMP *C)
{

  /*Extracts component c0 at level g i.e.
    Finds components that are below the extracted component.
    Returns list of all the children in those components.*/

  /*NOTE: Newer version extract_component2 is currently used.*/

  int i,ii,l,ci,c_up,c_down,lev;
  int **ind;
  int **components;
  int *ncomponents;

  if( (ncomponents = (int *) calloc(g+1,sizeof(int))) == NULL)
    alloc_error("extract_component");

  if( (components = (int **) malloc((g+1)*sizeof(int*))) == NULL)
    alloc_error("extract_component");

  for(i=0;i<g+1;i++)
    if( (components[i] = (int *) malloc(sizeof(int))) == NULL)
      alloc_error("extract_component");

  ncomponents[g]=1;
  components[g][0]=c0;


  /*This does not take clones into account i.e. those children that some parent may have on some lower levels.
    Only the highest level is checked here for each parent.
    Could be modified by using the offspring array...
  */

  for(l=g-1; l>-1; --l)
    {
      for(ci=0;ci<ncomponents[l+1];ci++)
	{
	  c_up=components[l+1][ci];
	  for(ii=0;ii < C[l+1].components[c_up].nchildren;ii++)
	    {
	      i=C[l+1].components[c_up].children[ii];
	      lev=C[l+1].parent_at_level[ordinal(i,C[l+1].individuals,C[l+1].n)][0];
	      
	      if(lev==-1) continue;

	      c_down=C[lev].p_comp[ordinal(i,C[lev].parents,C[lev].nparents)];
	      if(ordinal(c_down,components[lev],ncomponents[lev])==-1)
		{
		  if( (components[lev]=(int*) realloc(components[lev],(ncomponents[lev]+1)*sizeof(int)))==NULL)
		    alloc_error("extract_component");
		  components[lev][ncomponents[lev]++]=c_down;
		}
	    }
	}
    }

  /*for(l=0;l<g+1;l++)
    {
      printf("Level %d (%d components):\n",l,ncomponents[l]);
      for(ci=0;ci<ncomponents[l];ci++)
	printf("%d ",components[l][ci]);
      printf("\n");
      }*/

  if( (ind = (int **) malloc((g+2)*sizeof(int*))) == NULL)
    alloc_error("extract_component");

  for(l=0;l<g+1;l++)
    {
      i=0;

      for(ci=0;ci<ncomponents[l];ci++)
	{
	  c_down=components[l][ci];
	  nind[l] += C[l].components[c_down].nchildren;
	}

      if( (ind[l]=(int*) malloc(nind[l]*sizeof(int)))==NULL)
	alloc_error("extract_component");

      for(ci=0;ci<ncomponents[l];ci++)
	{
	  c_down=components[l][ci];
	  for(ii=0;ii<C[l].components[c_down].nchildren;ii++)
	    ind[l][i++]=C[l].components[c_down].children[ii];
	}
    }

  nind[g+1]=C[g].components[c0].nparents;
  if( (ind[g+1]=(int*) malloc(nind[g+1]*sizeof(int)))==NULL)
    alloc_error("extract_component");

  for(ii=0;ii<C[g].components[c0].nparents;ii++)
    ind[g+1][ii]=C[g].components[c0].parents[ii];
  

  for(l=0;l<g+2;l++)
    {
      printf("Level %d (%d individuals):\n",l,nind[l]);
      for(ii=0;ii<nind[l];ii++)
	{
	  printf("%d ",ind[l][ii]+1);
	  if(ii%15==14)
	    printf("\n");
	}
      printf("\n");
    }

  free(ncomponents);
  for(i=0;i<g+1;i++)
    free(components[i]);
  free(components);

  return ind;

}

int** extract_component2(int g, int c0, int *nind, COMP *C, CHILDREN *offspring, int *levels)
{

 /*Extracts component c0 at level g i.e.
   Finds all the descendants of the parents of the extracted component.
   Returns list that contains those parents and all the descendants.
   Descendants are in the arrays ind[0][] ... ind[g][] according to their levels.
   Parents of the component c0 are in the list ind[g+1][]
   (Note that not all the parents necessarily belong to the level g+1. They may be also at higher levels.)
*/
  

  int i,ii,l,ci,lev;
  int **ind;

   if( (ind = (int **) malloc((g+2)*sizeof(int*))) == NULL)
    alloc_error("extract_component2");

  for(l=0;l<g+1;l++)
    if( (ind[l]=(int*) malloc(sizeof(int)))==NULL)
      alloc_error("extract_component");

  nind[g+1]=C[g].components[c0].nparents;
  if( (ind[g+1]=(int*) malloc(nind[g+1]*sizeof(int)))==NULL)
    alloc_error("extract_component");

  for(ii=0;ii<C[g].components[c0].nparents;ii++)
    ind[g+1][ii]=C[g].components[c0].parents[ii];

  for(l=g+1;l>0;l--)
    {
      for(ii=0;ii<nind[l];ii++)
	{
	  for(ci=0;ci<offspring[ind[l][ii]].nchildren;ci++)
	    {
	      i=offspring[ind[l][ii]].children[ci];
	      lev=levels[i];
	      if(ordinal(i,ind[lev],nind[lev])==-1)
		{
		  if( (ind[lev]=(int*) realloc(ind[lev],(nind[lev]+1)*sizeof(int)))==NULL)
		    alloc_error("extract_component");
		  ind[lev][nind[lev]++]=i;
		}
	    }
	}
    }


  /*  for(l=0;l<g+2;l++)
    {
      printf("Level %d (%d individuals):\n",l,nind[l]);
      for(ii=0;ii<nind[l];ii++)
	printf("%d ",ind[l][ii]+1);
      printf("\n");
      }*/

  return ind;

}

int **extract(pair *extracted, int n_ex, int max_g, int *nind, COMP *C, CHILDREN *offspring, int *levels)
{

  int i,c,g,l,ii,ll;
  int *t_nind;
  int **t_ind,**ind;


  if( (ind = (int **) malloc((max_g+2)*sizeof(int*))) == NULL)
    alloc_error("extract_component");

  for(l=0;l<max_g+2;l++)
    if( (ind[l]=(int*) malloc(sizeof(int)))==NULL)
      alloc_error("extract_component");


  for(i=0;i<n_ex;i++)
    {
      g=extracted[i][0];
      c=extracted[i][1];

      if( (t_nind=(int*) calloc((g+2),sizeof(int)))==NULL)
	alloc_error("extract_component");

      t_ind=extract_component2(g,c,t_nind,C,offspring,levels);
      //t_ind=extract_component(g,c,t_nind,C);

      for(l=0;l<g+1;l++)
	{
	  for(ii=0;ii<t_nind[l];ii++)
	    if(ordinal(t_ind[l][ii],ind[l],nind[l])==-1)
	      {
		if( (ind[l]=(int*) realloc(ind[l],(nind[l]+1)*sizeof(int)))==NULL)
		  alloc_error("extract_component");
		ind[l][nind[l]++]=t_ind[l][ii];
	      }
	}

      for(ii=0;ii<t_nind[g+1];ii++)
	{
	  l=(levels[t_ind[g+1][ii]]<max_g+1)?levels[t_ind[g+1][ii]]:max_g+1;
	  if(ordinal(t_ind[g+1][ii],ind[l],nind[l])==-1)
	    {
	      if( (ind[l]=(int*) realloc(ind[l],(nind[l]+1)*sizeof(int)))==NULL)
		alloc_error("extract_component");
	      ll=(l>g+1)?g+1:l;
	      ind[l][nind[l]++]=t_ind[ll][ii];
	    }
	}

      free(t_nind);
      
    }
  
  i=0;
  for(l=0;l<max_g+2;l++)
    {
      printf("Level %d (%d individuals):\n",l,nind[l]);
      for(ii=0;ii<nind[l];ii++)
	{
	  printf("%d ",ind[l][ii]+1);
	  i++;
	  if(ii%15==14)
	    printf("\n");
	}
      printf("\n");
    }

  /*This is used to write the indexes of the individuals in extracted components or below to the file*/
  /*{
    FILE *color=fopen("color_these","w");
    fprintf(color,"%d\n",i);
    for(l=0;l<max_g+2;l++)
      for(ii=0;ii<nind[l];ii++)
	fprintf(color,"%d\n",ind[l][ii]+1);
    fclose(color);
    }*/

  return ind;

}




void undefine_rest(int *nind, int **ind, pair **genes, int marker, int *levels, int g)
{
  int i;

  for(i=0;i<N_individuals;i++)
    {
      if(genes[i][marker][0]+genes[i][marker][1] > 0)
	{
	  if( (levels[i]>=g+1 && ordinal(i,ind[g+1],nind[g+1])==-1)  || (levels[i] < g+1 && ordinal(i,ind[levels[i]],nind[levels[i]])==-1) )
	    {
	      genes[i][marker][0]=0;
	      genes[i][marker][1]=0;
	    }
	}
    }
}



void print_extracted_component(FILE* out,COMP *C,pair **genes, int g,int c0,int markers)
{

  int i,p,m;

  fprintf(out,"\nPrinting data of component %d at level %d.\n\n",c0,g);
  fprintf(out,"Parents (%d):",C[g].components[c0].nparents);

 
      for(i=0;i<C[g].components[c0].nparents;i++)
	{
	  p=C[g].components[c0].parents[i];

	  fprintf(out,"\n%d:",p+1);
	  for(m=0;m<markers;m++)
	    fprintf(out,"(%d,%d)",genes[p][m][0],genes[p][m][1]);
	}


      fprintf(out,"\nChildren (%d):",C[g].components[c0].nchildren);

 
      for(i=0;i<C[g].components[c0].nchildren;i++)
	{
	  p=C[g].components[c0].children[i];

	  fprintf(out,"\n%d:",p+1);
	  for(m=0;m<markers;m++)
	    fprintf(out,"(%d,%d)",genes[p][m][0],genes[p][m][1]);
	}

      fprintf(out,"\nComponent data ENDS.\n");
}



void print_extracted(FILE* out,COMP *C,pair **genes, pair *extracted, int n_ext,int markers)
{
  int i;

  for(i=0;i<n_ext;i++)
    print_extracted_component(out,C,genes,extracted[i][0],extracted[i][1],markers);
}

void print_suggestions(FILE* out,COMP *C,pair *extracted, int n_ext)
{
  int i,j,g,c,p,count=0;

  for(j=0;j<n_ext;j++)
    {
      g=extracted[j][0];
      c=extracted[j][1];
      if(g>0)
	{
	  fprintf(out,"Components below [%d,%d] are: ",g,c);
	  for(i=0;i<C[g].components[c].nchildren;i++)
	    {
	      if((p=ordinal(C[g].components[c].children[i],C[g-1].parents,C[g-1].nparents))> -1)
		{
		  fprintf(out,"[%d,%d] ",g-1,C[g-1].p_comp[p]);
		  count++;
		}
	      if(count%15==14)
		fprintf(out,"\n");
	    }
	  fprintf(out,"\n");
	}

    }
}




void write_new_datafile(FILE *data, pair **genes, COMP* C, int *knowndata, pair *extracted, int n_ext)
{

  /*This adds data of parents of every extracted component to 
    the genes and writes them to the EXTRACTED_DATA_FILE.*/

  int i,j,p,m,comp,lev,ext;
  int sum=N_knowndata;
  FILE *new;
  pair original[N_markers];

  for(ext=0;ext<n_ext;ext++)
    {
      lev=extracted[ext][0];
      comp=extracted[ext][1];
      for(i=0;i<C[lev].components[comp].nparents;i++)
	if(ordinal(i,C[lev].components[comp].knowndata,C[lev].components[comp].nknowndata)==-1)
	  sum++;
    }

  fscanf(data,"%d",&i);
  
  for(;i>0;i--)
    {
      fscanf(data,"%d\n",&p);
      for(j=0;j<2;j++)
	for(m=0;m<N_markers;m++)
	  fscanf(data,"%d ",&original[m][j]);

      for(m=0;m<N_markers;m++)
	if(original[m][0]>0 && original[m][1]>0)
	  {genes[p-1][m][0]=original[m][0];genes[p-1][m][1]=original[m][1];}
	else
	  if(!compatible(original[m][0],genes[p-1][m]) || !compatible(original[m][1],genes[p-1][m]))
	    {
	      fprintf(stderr,"ERROR in write_new_datafile with individual %d!\n",p);
	      exit(5);
	    }
    }

  fclose(data);

  new=fopen(EXTRACTED_DATA_FILE,"w"); //the new datafile, original is not modified!
  fprintf(new,"%d",sum);

  for(i=0;i<N_knowndata;i++)
    {
      p=knowndata[i];
      fprintf(new,"\n\n%d",p+1);
      for(j=0;j<2;j++)
	{
	  fprintf(new,"\n");
	  for(m=0;m<N_markers;m++)
	    fprintf(new,"%d ",genes[p][m][j]);
	}
    }

  for(ext=0;ext<n_ext;ext++)
    {
      lev=extracted[ext][0];
      comp=extracted[ext][1];    
      for(i=0;i<C[lev].components[comp].nparents;i++)
	{
	  p=C[lev].components[comp].parents[i];
	  if(ordinal(i,C[lev].components[comp].knowndata,C[lev].components[comp].nknowndata)==-1)
	    {
	      fprintf(new,"\n\n%d",p+1);
	      for(j=0;j<2;j++)
		{
		  fprintf(new,"\n");
		  for(m=0;m<N_markers;m++)
		    fprintf(new,"%d ",genes[p][m][j]);
		}
	    }
	}
    }
  fclose(new);

}


int write_new_datafile2(FILE *data, pair **genes, int *knowndata, int max_g, int **ind, int* nind, pair *extracted, int n_ext, COMP *C)
{

  /*This adds data of people on the highest level in ind 
    that are not yet included in the knowndata and that have defined genes.
    In addition all the defined alleles of parents at extracted components are fixed.
    Writes to the EXTRACTED_DATA_FILE.
    (Difference with write_new_datafile is that this (usually) 
    writes something even though parents in extracted components have
    undefined genotypes.)*/

  int i,j,p,m,g,ext,lev,comp;
  FILE *new;
  pair original[N_markers];
  int *add;
  int nadd=0;

  i=0;
  for(g=0;g<max_g+2;g++)
    i += nind[g];

  if( (add = (int*) malloc(i*sizeof(int)))==NULL)
    alloc_error("write_new_datafile2");

  printf("\nData added for:");
  for(g=max_g+1;g>=0;g--)
    {
      for(i=0;i<nind[g];i++)
	if(ordinal(ind[g][i],knowndata,N_knowndata)==-1)
	  {
	    for(m=0;m<N_markers;m++)
	      if(genes[ind[g][i]][m][0] + genes[ind[g][i]][m][1]>0)
		{
		  add[nadd++]=ind[g][i];
		  break;
		}
	  }
      if(nadd) break;
    }
  //printf("before componentwise %d\n",nadd);
  for(ext=0;ext<n_ext;ext++)
    {
      lev=extracted[ext][0];
      comp=extracted[ext][1];    
      for(i=0;i<C[lev].components[comp].nparents;i++)
	{
	  p=C[lev].components[comp].parents[i];
	  if(ordinal(i,C[lev].components[comp].knowndata,C[lev].components[comp].nknowndata)==-1 && ordinal(p,add,nadd)==-1 )
	    {
	      for(m=0;m<N_markers;m++)
		if(genes[p][m][0] + genes[p][m][1]>0)
		  {
		    add[nadd++]=p;
		    break;
		  }
	    }
	}
    }

  fscanf(data,"%d",&i);
  
  for(;i>0;i--) // read the original data from datafile
    {
      fscanf(data,"%d\n",&p);
      for(j=0;j<2;j++)
	for(m=0;m<N_markers;m++)
	  fscanf(data,"%d ",&original[m][j]);

      for(m=0;m<N_markers;m++)
	if(original[m][0]>0 && original[m][1]>0)
	  {genes[p-1][m][0]=original[m][0];genes[p-1][m][1]=original[m][1];}
	else
	  if(!compatible(original[m][0],genes[p-1][m]) || !compatible(original[m][1],genes[p-1][m]))
	    {
	      fprintf(stderr,"ERROR in write_new_datafile with individual %d!\n",p);
	      exit(5);
	    }
    }

  fclose(data);

  new=fopen(EXTRACTED_DATA_FILE,"w"); //the new datafile, original is not modified!
  fprintf(new,"%d",N_knowndata+nadd);

  for(i=0;i<N_knowndata;i++)
    {
      p=knowndata[i];
      fprintf(new,"\n\n%d",p+1);
      for(j=0;j<2;j++)
	{
	  fprintf(new,"\n");
	  for(m=0;m<N_markers;m++)
	    fprintf(new,"%d ",genes[p][m][j]);
	}
    }

  for(i=0;i<nadd;i++)
    {
      fprintf(new,"\n\n%d",add[i]+1);
      printf("\n%d:",add[i]+1);
      for(j=0;j<2;j++)
	{
	  fprintf(new,"\n");
	  for(m=0;m<N_markers;m++)
	    {
	      fprintf(new,"%d ",genes[add[i]][m][j]);
	      printf("%d ",genes[add[i]][m][j]);
	    }
	  printf(";");
	}
    }
  printf("\nThat is %d individuals.\n",nadd);
  fclose(new);
  free(add);

  return nadd;
}


void read_component(FILE *compfile, int *l, int *c,COMP *C)
{
  char label[40];

  fscanf(compfile,"%s\n %d\n",label,l);
  fscanf(compfile,"%s\n %d\n",label,c);

  if(*l>N_levels-2 || *l<0)
    {
      fprintf(stderr,"ERROR: Level %d given in file \"extracted_component\" is invalid (highest possible=%d).\n",*l,N_levels-2);
      exit(3);
    }
  if(*c>C[*l].ncomponents-1 || *c<0)
    {
      fprintf(stderr,"ERROR: Component %d given in file \"extracted_component\" is invalid (highest possible at level %d = %d).\n",*c,*l,C[*l].ncomponents-1);
      exit(3);
    }

}


pair *read_extracted(FILE *ext,int *n_ext,int *max_ext_level, COMP *C)
{
  
  int status,i,l,c;
  pair *extracted;
  
  if( (extracted=(pair*) malloc(sizeof(pair)))==NULL)
    alloc_error("read_extracted with extracted");

  *n_ext=0;
  *max_ext_level=0;

  do
    {
      status=fscanf(ext,"%d %d",&l,&c);
      if(status != EOF) 
	{
	  (*n_ext)++;
	  if( (extracted=(pair*) realloc(extracted,(*n_ext)*sizeof(pair)))==NULL)
	     exit(1);
	   extracted[*n_ext-1][0]=l;
	   extracted[*n_ext-1][1]=c;
	   *max_ext_level=(l>(*max_ext_level))?l:(*max_ext_level);
	}
    }while(status != EOF);

  for(i=0;i<*n_ext;i++)
    {
      l=extracted[i][0];
      c=extracted[i][1];
      if(l>N_levels-2 || l<0)
	{
	  fprintf(stderr,"ERROR: Level %d given in file \"extracted_component\" is invalid (highest possible=%d).\n",l,N_levels-2);
	  exit(3);
	}
      if(c>C[l].ncomponents-1 || c<0)
	{
	  fprintf(stderr,"ERROR: Component %d given in file \"extracted_component\" is invalid (highest possible at level %d = %d).\n",c,l,C[l].ncomponents-1);
	  exit(3);
	}
    }

  fprintf(stdout,"Extracting %d components: ",*n_ext);
  for(i=0;i<*n_ext;i++)
    fprintf(stdout,"[%d,%d] ",extracted[i][0],extracted[i][1]);
  fprintf(stdout,"\n");

  return extracted;
}


int find_extracted_configurations(g_list *G, COMP *C, int g, int **fixed_parents, pair** genes, int marker, double ***conditionalF,int **ind, int *nind,int ex_level, int *levels)
{

  int t,i,j,l,k,h;
  int nchildren,component,Father,Mother,ntensor,ntensor0,nparents,child,father,mother;
  pair *ctensor,**(ttensor);
  double *proba,sumproba;

  nchildren=C[g].n;

  if( (ctensor = (pair*) malloc(nchildren*sizeof(pair))) == NULL)
    alloc_error("find_extracted_configurations");

  for(i=0; i < nchildren; i++)
    {
      l=C[g].parent_at_level[i][0];
      if(l==-1)
	{
	  ctensor[i][0]=genes[C[g].individuals[i]][marker][0];
	  ctensor[i][1]=genes[C[g].individuals[i]][marker][1];
	}
      else
	{
	  t=C[g].parent_at_level[i][1];
	  k=C[l].p_comp[t];
	  h=C[l].pind_in_comp[t];
	  
	  ctensor[i][0] = G[l].comp[k].tensor[G[l].ordered[G[l+1].prev_config].component[k]][h][0];
	  ctensor[i][1] = G[l].comp[k].tensor[G[l].ordered[G[l+1].prev_config].component[k]][h][1];
	}
    }

  if( (G[g].nconfig = (int *) malloc(C[g].ncomponents*sizeof(int))) == NULL)
    alloc_error("find_extracted_configurations");

  if( (G[g].comp = (c_list *) malloc(C[g].ncomponents*sizeof(c_list))) == NULL)
    alloc_error("find_extracted_configurations");


  for(component=0; component < C[g].ncomponents; component++)
    {

      nchildren = C[g].components[component].nchildren;
      nparents = C[g].components[component].nparents;

      if( (ttensor = (pair**) malloc( sizeof(pair*))) == NULL)
	alloc_error("find_extracted_configurations");

      if( (ttensor[0] = (pair*) calloc(nparents, sizeof(pair))) == NULL)
	alloc_error("find_extracted_configurations");

     
      ntensor=1;

      for(j=0; j<C[g].components[component].nknowndata; j++)
	{
	  k=C[g].components[component].knowndata[j];
	  t=C[g].components[component].parents[k];

	  ttensor[0][k][0]=genes[t][marker][0];
	  ttensor[0][k][1]=genes[t][marker][1];
	}

      for(j=0; j<C[g].components[component].ncloned; j++)
	{
	  t=C[g].components[component].parents[C[g].components[component].cloned[j][0]];
	  l=C[g].components[component].cloned[j][1];
	  h=ordinal(t,C[l].parents,C[l].nparents);
	  k=C[l].p_comp[h];
	  t=C[g].components[component].cloned[j][0];

	  ttensor[0][t][0]=G[l].comp[k].tensor[G[l].ordered[G[l+1].prev_config].component[k]][C[l].pind_in_comp[h]][0];
	  ttensor[0][t][1]=G[l].comp[k].tensor[G[l].ordered[G[l+1].prev_config].component[k]][C[l].pind_in_comp[h]][1];

	}

      for(j=0;j<nchildren ;j++)
	{

	  child=C[g].components[component].children[j];
	  t= ordinal(child,C[g].individuals,C[g].n);

	  Father = fixed_parents[child][0];
	  father = ordinal(Father,C[g].components[component].parents,C[g].components[component].nparents);

	  Mother = fixed_parents[child][1];
	  mother = ordinal(Mother,C[g].components[component].parents,C[g].components[component].nparents);

	  ntensor0=ntensor;

	  for(i=0;i<ntensor0;i++)
	    ttensor=update_list(ctensor[t],father,Father,mother,Mother,ttensor,&i,&ntensor,&ntensor0, conditionalF,nparents,N_alleles[marker]);

	  for(i=0;i<ntensor;i++)
	    join_identical2(ttensor,&i,&ntensor,nparents);
	      
	}
    
      if(ntensor>0)
	{

	  if ( (proba = (double *) malloc(ntensor*sizeof(double))) == NULL)
	    alloc_error("find_extracted_configurations");

	  for(i=0;i<ntensor;i++)
	    {

	      proba[i]=1.0;
	      for(j=0;j<nparents;j++)
		{
		  k=C[g].components[component].parents[j];
		  proba[i] *= conditional_probability(k,ttensor[i][j],conditionalF,N_alleles[marker]);
		}
	    }
	  order_component(ttensor,proba,ntensor);

	  /*This is the new part for extraction procedure.*/ 

	  if(g <= ex_level)
	    {
	      for(j=0;j<nparents;j++)
		{
		  t=C[g].components[component].parents[j];
		  k=(levels[t]>ex_level)?ex_level+1:levels[t];
		  if(ordinal(t,ind[k],nind[k])==-1 && nind[k]>0)
		    {
		      for(i=0;i<ntensor;i++)
			{
			  ttensor[i][j][0]=0;
			  ttensor[i][j][1]=0;
			}
		    }
		}

	      for(i=0;i<ntensor;i++)
		{
		  j=i;
		  join_identical2(ttensor,&i,&ntensor,nparents);
		  if(i<j)//original ttensor[i] has been deleted 
		    proba[j]=proba[ntensor];
		}

	      if( (proba = (double*) realloc(proba,ntensor*sizeof(double))) == NULL)
		alloc_error("find_extracted_configurations");
	    }
	  /*New part ends here.*/

	  ntensor0=ntensor;
	  if(M_comp>0 && ntensor>=M_comp)
	    ntensor = M_comp;

	  G[g].nconfig[component]=ntensor;

	  if( ((G[g].comp[component].tensor) = (pair**) malloc(ntensor * sizeof(pair*))) == NULL)
	    alloc_error("find_extracted_configurations");

	  sumproba=0.0;

	  for(i=0;i<ntensor;i++)
	    {
	      if( (G[g].comp[component].tensor[i] = (pair*) malloc(nparents* sizeof(pair))) == NULL)
		exit(1);

	      sumproba += proba[i];
	    }

	  if( ((G[g].comp[component].proba) = (double*) malloc(ntensor * sizeof(double))) == NULL)
	    alloc_error("find_extracted_configurations");

	  for(i=0;i<ntensor;i++)
	    {   
	      G[g].comp[component].proba[i]=proba[i]/sumproba;
	      
	      for(j=0;j<nparents;j++)
		{
		  G[g].comp[component].tensor[i][j][0]=ttensor[i][j][0];
		  G[g].comp[component].tensor[i][j][1]=ttensor[i][j][1];
		}
	      free(ttensor[i]);
	    }

	  for(i=ntensor;i<ntensor0;i++)
	    free(ttensor[i]);

	  free(proba);
	  free(ttensor);
	}
      else
	{   /* no consistent configurations found.*/
	
	  free(ttensor);
	  free(ctensor);

	  for(i=0;i<component;i++)
	    {
	      for(j=0;j<G[g].nconfig[i];j++)
		free(G[g].comp[i].tensor[j]);

	      free(G[g].comp[i].tensor);
	      free(G[g].comp[i].proba);

	    }

	  free(G[g].nconfig);
	  free(G[g].comp);

	  return component;
	
	}

    }

  G[g].n=0; //not yet defined, since may become too large
  G[g].log_n=0;

  for(component=0;component<C[g].ncomponents; component++)
    G[g].log_n += log(G[g].nconfig[component]);      

  free(ctensor);
  //  printf("%f\n",exp(G[g].log_n));

  return -1;
  
}  






